@@ -1,119 +1,15 @@
+# SPDX-License-Identifier: GPL-2.0-or-later
+#
 # bash tab completion for the nvme command line utility
 # (unfortunately, bash won't let me add descriptions to cmds)
 # Kelly Kaoudis kelly.n.kaoudis at intel.com, Aug. 2015
 
-# Constant to indicate command has no options
-readonly NO_OPTS=""
-
-# Associative array of plugins and associated subcommands
-# Order here is same as PLUGIN_OBJS in Makefile
-typeset -A _plugin_subcmds
-readonly _plugin_subcmds=(
-	[intel]="id-ctrl internal-log lat-stats \
-		set-bucket-thresholds lat-stats-tracking \
-		market-name smart-log-add temp-stats"
-	[amzn]="id-ctrl"
-	[lnvm]="list info id-ns chunk-log init create \
-		remove factory diag-bbtbl diag-set-bbtbl"
-	[memblaze]="smart-log-add get-pm-status set-pm-status \
-		select-download lat-stats lat-stats-print lat-log \
-		lat-log-print clear-error-log"
-	[wdc]="cap-diag drive-log get-crash-dump get-pfail-dump \
-		id-ctrl purge purge-monitor vs-internal-log \
-		vs-nand-stats vs-smart-add-log clear-pcie-correctable-errors \
-		drive-essentials get-drive-status clear-assert-dump \
-		drive-resize vs-fw-activate-history clear-fw-activate-history \
-		enc-get-log vs-telemetry-controller-option \
-		vs-error-reason-identifier log-page-directory \
-		namespace-resize vs-drive-info vs-temperature-stats \
-		capabilities cloud-SSD-plugin-version vs-pcie-stats"
-	[huawei]="list id-ctrl"
-	[netapp]="smdevices ontapdevices"
-	[toshiba]="vs-smart-add-log vs-internal-log \
-		clear-pcie-correctable-errors"
-	[micron]="select-download vs-temperature-stats vs-pcie-stats \
-		clear-pcie-correctable-errors vs-internal-log \
-		vs-telemetry-controller-option vs-nand-stats \
-		vs-drive-info plugin-version cloud-SSD-plugin-version \
-		log-page-directory vs-fw-activate-history \
-		vs-error-reason-identifier vs-smart-add-log \
-		clear-fw-activate-history vs-smbus-option"
-	[seagate]="vs-temperature-stats vs-log-page-sup \
-		vs-smart-add-log vs-pcie-stats clear-pcie-correctable-errors \
-		get-host-tele get-ctrl-tele vs-internal-log \
-		plugin-version"
-	[virtium]="save-smart-to-vtview-log show-identify"
-	[shannon]="smart-log-add get-feature-add set-feature-add id-ctrl"
-	[dera]="smart-log-add"
-	[sfx]="smart-log-add lat-stats get-bad-block query-cap \
-		change-cap set-feature get-feature"
-	[transcend]="healthvalue badblock"
-	[zns]="id-ctrl id-ns zone-mgmt-recv \
-		zone-mgmt-send report-zones close-zone \
-		finish-zone open-zone reset-zone offline-zone \
-		set-zone-desc zone-append changed-zone-list"
-	[nvidia]="id-ctrl"
-	[ymtc]="smart-log-add"
-)
-
-# Associative array mapping plugins to coresponding option completions
-typeset -A _plugin_funcs
-readonly _plugin_funcs=(
-	[intel]="plugin_intel_opts"
-	[amzn]="plugin_amzn_opts"
-	[lnvm]="plugin_lnvm_opts"
-	[memblaze]="plugin_memblaze_opts"
-	[wdc]="plugin_wdc_opts"
-	[huawei]="plugin_huawei_opts"
-	[toshiba]="plugin_toshiba_opts"
-	[micron]="plugin_micron_opts"
-	[seagate]="plugin_seagate_opts"
-	[virtium]="plugin_virtium_opts"
-	[shannon]="plugin_shannon_opts"
-	[dera]="plugin_dera_opts"
-	[sfx]="pluginx_sfx_opts"
-	[transcend]="plugin_transcend_opts"
-	[zns]="plugin_zns_opts"
-	[nvidia]="plugin_nvidia_opts"
-	[ymtc]="plugin_ymtc_opts"
-)
-
-# Top level commands
-_cmds="list list-subsys id-ctrl id-ns \
-	id-ns-granularity list-ns list-ctrl \
-	nvm-id-ctrl primary-ctrl-caps list-secondary \
-	ns-descs id-nvmset id-uuid id-iocs create-ns \
-	delete-ns get-ns-id get-log telemetry-log \
-	fw-log changed-ns-list-log smart-log ana-log \
-	error-log effects-log endurance-log \
-	predictable-lat-log pred-lat-event-agg-log \
-	persistent-event-log endurance-agg-log \
-	lba-status-log resv-notif-log get-feature \
-	device-self-test self-test-log set-feature \
-	set-property get-property format fw-commit \
-	fw-download admin-passthru io-passthru \
-	security-send security-recv get-lba-status \
-	resv-acquire resv-register resv-release \
-	resv-report dsm copy flush compare read \
-	write write-zeros write-uncor verify \
-	sanitize sanitize-log reset subsystem-reset \
-	ns-rescan show-regs discover connect-all \
-	connect disconnect disconnect-all gen-hostnqn \
-	show-hostnqn dir-receive dir-send virt-mgmt \
-	rpmb boot-part-log fid-support-effects-log \
-	supported-log-pages lockdown"
-	supported-log-pages list-endgrp"
-
-# Add plugins:
-for plugin in "${!_plugin_subcmds[@]}"; do
-	_cmds+=" $plugin"
-done
-
-cmds+=" version help"
-
 nvme_list_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
+	local vals=""
+	local opt=""
+	local val=""
 
 	local nonopt_args=0
 	for (( i=0; i < ${#words[@]}-1; i++ )); do
@@ -127,6 +23,19 @@ nvme_list_opts () {
 	fi
 
 	opts+=" "
+	vals+=" "
+
+	if [[ $cur != -* ]] && [[ $cur != "" ]] && [[ $prev == "=" ]] && [[ ${words[$cword-2]} == --* ]]; then
+		opt+="${words[$cword-2]}"
+		val+="$cur"
+	elif [[ $cur == "" ]] && [[ $prev != "=" ]] || [[ $cur == "=" ]] && [[ $prev == --* ]]; then
+		opt+="$prev"
+	elif [[ $cur != "=" ]] && [[ $prev != --* ]] && [[ $prev != "=" ]]; then
+		opt+="$prev"
+		val+="$cur"
+	else
+		opt+="$cur"
+	fi
 
 	# Listed here in the same order as in nvme-builtin.h
 	case "$1" in
@@ -138,19 +47,23 @@ nvme_list_opts () {
 			;;
 		"id-ctrl")
 		opts+=" --raw-binary -b --human-readable -H \
-			--vendor-specific -v --output-format= -o"
+			--vendor-specific -V --output-format= -o"
 			;;
 		"id-ns")
 		opts+=" --namespace-id= -n --raw-binary -b \
-			--human-readable -H --vendor-specific -v \
+			--human-readable -H --vendor-specific -V \
 			--force -f --output-format= -o"
 			;;
 		"id-ns-granularity")
 		opts+=" --output-format= -o"
 			;;
+		"id-ns-lba-format")
+		opts+=" --lba-format-index= -i --uuid-index= -U \
+			--verbose -v --output-format= -o"
+			;;
 		"list-ns")
 		opts+=" --namespace-id= -n --al -a --csi= -y \
-			--outputformat= -o"
+			--outputformat= -o --timeout= -t"
 			;;
 		"list-ctrl")
 		opts+=" --namespace-id= -n --cntid= -c \
@@ -160,12 +73,17 @@ nvme_list_opts () {
 		opts+=" --namespace-id= -n --raw-binary -b \
 			--human-readable -H --output-format= -o"
 			;;
-		"id-domain")
-		opts+=" --dom-id= -d --output-format= -o"
-			;;
 		"nvm-id-ctrl")
 		opts+=" --output-format= -o"
 			;;
+		"nvm-id-ns")
+		opts+=" --namespace-id= -n --uuid-index= -U\
+			--verbose -v --output-format= -o"
+			;;
+		"nvm-id-ns-lba-format")
+		opts+=" --lba-format-index= -i --uuid-index= -U \
+			--verbose -v --output-format= -o"
+			;;
 		"primary-ctrl-caps")
 		opts+=" --output-format= -o --human-readable -H"
 			;;
@@ -186,12 +104,17 @@ nvme_list_opts () {
 		opts+=" --endgrp-id= -i --output-format= -o"
 			;;
 		"id-iocs")
-		opts+=" --controller-id= -c --output-format= -o --human-readable -H"
+		opts+=" --controller-id= -c"
+			;;
+		"id-domain")
+		opts+=" --domain-id= -c --output-format= -o"
 			;;
 		"create-ns")
 		opts+=" --nsze= -s --ncap= -c --flbas= -f \
-			--dps= -d --nmic= -n --anagrp-id= -a --nvmset-id= -i \
-			--block-size= -b --timeout= -t--csi= -y"
+			--dps= -d --nmic= -m --anagrp-id= -a --nvmset-id= -i \
+			--block-size= -b --timeout= -t --csi= -y --lbstm= -l \
+			--nphndls= -n --nsze-si= -S --ncap-si= -C --azr -z --rar= -r \
+			--ror= -O --rnumzrwa= -u --phndls= -p --endg-id= -e"
 			;;
 		"delete-ns")
 		opts+=" -namespace-id= -n --timeout= -t"
@@ -207,7 +130,7 @@ nvme_list_opts () {
 			;;
 		"get-log")
 		opts+=" --log-id= -i --log-len= -l --namespace-id= -n \
-			--aen= -a --lpo= -o --lsp= -s --lsi= -S \
+			--aen= -a --lpo= -O --lsp= -s --lsi= -S \
 			--rae -r --uuid-index= -U --csi= -y --ot -O \
 			--raw-binary -b"
 			;;
@@ -215,7 +138,7 @@ nvme_list_opts () {
 		opts+=" --output-format= -o --human-readable -H"
 			;;
 		"telemetry-log")
-		opts+=" --output-file= -o --host-generate= -g \
+		opts+=" --output-file= -O --host-generate= -g \
 			--controller-init -c --data-area= -d"
 			;;
 		"fw-log")
@@ -226,7 +149,7 @@ nvme_list_opts () {
 			;;
 		"smart-log")
 		opts+=" --namespace-id= -n --raw-binary -b \
-			--output-format= -o"
+			--output-format= -o --verbose -v"
 			;;
 		"ana-log")
 		opts+=" --output-format -o"
@@ -240,7 +163,7 @@ nvme_list_opts () {
 			;;
 		"effects-log")
 		opts+=" --output-format= -o --human-readable -H \
-			--raw-binary -b"
+			--raw-binary -b --timeout= -t"
 			;;
 		"endurance-log")
 		opts+=" --output-format= -o --group-id -g"
@@ -250,7 +173,7 @@ nvme_list_opts () {
 			--output-format= -o"
 			;;
 		"pred-lat-event-agg-log")
-		opts+=" --log-entries= -e  --rae -r \
+		opts+=" --log-entries= -e --rae -r \
 			--raw-binary -b --output-format= -o"
 			;;
 		"persistent-event-log")
@@ -258,7 +181,7 @@ nvme_list_opts () {
 			--raw-binary -b --output-format= -o"
 			;;
 		"endurance-event-agg-log")
-		opts+=" --log-entries= -e  --rae -r \
+		opts+=" --log-entries= -e --rae -r \
 			--raw-binary -b --output-format= -o"
 			;;
 		"lba-status-log")
@@ -271,13 +194,21 @@ nvme_list_opts () {
 		opts+=" --lsp -s --output-file= -f \
 			--output-format= -o"
 			;;
+		"media-unit-stat-log")
+		opts+=" --dom-id= -d --output-format= -o \
+			--raw-binary -b"
+			;;
+		"supported-cap-config-log")
+		opts+=" --dom-id= -d --output-format= -o \
+				--raw-binary -b"
+			;;
 		"get-feature")
 		opts+=" --namespace-id= -n --feature-id= -f --sel= -s \
 			--data-len= -l --cdw11= --c -uuid-index= -U --raw-binary -b \
-			--human-readable -H"
+			--human-readable -H --timeout= -t --changed -C"
 			;;
 		"device-self-test")
-		opts+=" --namespace-id= -n --self-test-code= -s"
+		opts+=" --namespace-id= -n --self-test-code= -s --timeout= -t"
 			;;
 		"self-test-log")
 		opts+=" --dst-entries= -e --output-format= -o \
@@ -286,34 +217,34 @@ nvme_list_opts () {
 		"set-feature")
 		opts+=" --namespace-id= -n --feature-id= -f --value= -v \
 			--data-len= -l -data= -d --value= -v --save -s --uuid-index= -U \
-			--cdw12= -c"
+			--cdw12= -c --timeout= -t"
 			;;
 		"set-property")
-		opts+=" --offset= -o --value= -v"
+		opts+=" --offset= -O --value= -V --timeout= -t"
 			;;
 		"get-property")
-		opts=+" --offset= -o --human-readable -H"
+		opts=+" --offset= -O --human-readable -H --timeout= -t"
 			;;
 		"format")
 		opts+=" --namespace-id= -n --timeout= -t --lbaf= -l \
 			--ses= -s --pil= -p -pi= -i --ms= -m --reset -r"
 			;;
 		"fw-commit")
-		opts+=" --slot= -s --action= -a --bpid= -b"
+		opts+=" --slot= -s --action= -a --bpid= -b --timeout= -t"
 			;;
 		"fw-download")
-		opts+=" --fw= -f --xfer= -x --offset= -o"
+		opts+=" --fw= -f --xfer= -x --offset= -O --timeout= -t"
 			;;
 		"capacity-mgmt")
-		opts+=" --operation= -f --element-id= -i --cap-lower= -l \
-				--cap-upper= -u"
+		opts+=" --operation= -O --element-id= -i --cap-lower= -l \
+			--cap-upper= -u --timeout= -t"
 			;;
 		"lockdown")
-		opts+=" --ofi= -O --ifc= -F --prhbt= -P \
-			-scp= -S --uuid -U"
+		opts+=" --ofi= -O --ifc= -f --prhbt= -p --scp= -s --uuid -U \
+			--timeout= -t"
 			;;
 		"admin-passthru")
-		opts+=" --opcode= -o --flags= -f --prefil= -p --rsvd= -R \
+		opts+=" --opcode= -O --flags= -f --prefil= -p --rsvd= -R \
 			--namespace-id= -n --data-len= -l --metadata-len= -m \
 			--timeout= -t --cdw2= -2 --cdw3= -3 --cdw10= -4 \
 			--cdw11= -5 --cdw12= -6 --cdw13= -7 --cdw14= -8 \
@@ -322,7 +253,7 @@ nvme_list_opts () {
 			--latency -T"
 			;;
 		"io-passthru")
-		opts+=" --opcode= -o --flags= -f --prefill= -p --rsvd= -R \
+		opts+=" --opcode= -O --flags= -f --prefill= -p --rsvd= -R \
 			--namespace-id= -n --data-len= -l --metadata-len= -m \
 			--timeout= -t --cdw2= -2 --cdw3= -3 --cdw10= -4 \
 			--cdw11= -5 --cdw12= -6 --cdw13= -7 --cdw14= -8 \
@@ -332,11 +263,11 @@ nvme_list_opts () {
 			;;
 		"security-send")
 		opts+=" --namespace-id= -n --file= -f --nssf= -N --secp= -p \
-			--spsp= -s --tl= -t"
+			--spsp= -s --tl= -t --timeout="
 			;;
 		"security-recv")
 		opts+=" --namespace-id= -n --size= -x --secp= -p --spsp= -s \
-			--al= -t --raw-binary -b"
+			--al= -t --raw-binary -b --timeout="
 			;;
 		"get-lba-status")
 		opts+=" --namespace-id= -n --start-lba= -s --max-dw= -m \
@@ -345,23 +276,24 @@ nvme_list_opts () {
 			;;
 		"resv-acquire")
 		opts+=" --namespace-id= -n --crkey= -c --prkey= -p \
-			--rtype= -t --racqa= -a --iekey= -i"
+			--rtype= -t --racqa= -a --iekey= -i --timeout="
 			;;
 		"resv-register")
 		opts+=" --namespace-id= -n --crkey= -c --nrkey= -k \
-			--rrega= -r --cptpl= -p --iekey -i"
+			--rrega= -r --cptpl= -p --iekey -i --timeout= -t"
 			;;
 		"resv-release")
 		opts+=" --namespace-id= -n --crkey -c --rtype= -t \
-			--rrela= -a --iekey -i"
+			--rrela= -a --iekey -i --timeout="
 			;;
 		"resv-report")
-		opts+=" --namespace-id= -n --numd= -d --cdw11 -c \
-			--output-format= -o --raw-binary -b"
+		opts+=" --namespace-id= -n --numd= -d --eds -e \
+			--raw-binary= -b --output-format= -o --timeout= -t"
 			;;
 		"dsm")
 		opts+=" --namespace-id= -n --ctx-attrs= -a --blocks= -b \
-			--slbs= -s --ad -d --idw -w --idr -r --cdw11= -c"
+			--slbs= -s --ad -d --idw -w --idr -r --cdw11= -c \
+			--timeout= -t"
 			;;
 		"copy")
 		opts+=" --namespace-id= -n --sdlba= -d --blocks= -b --slbs= -s \
@@ -370,59 +302,68 @@ nvme_list_opts () {
 			--ref-tag= -r --expected-ref-tag= -R \
 			--app-tag= -a --expected-app-tag= -A \
 			--app-tag-mask= -m --expected-app-tag-mask= -M \
-			--dir-type= -T --dir-spec= -S --format= -F"
+			--dir-type= -T --dir-spec= -S --format= -F --timeout= -t"
 			;;
 		"flush")
 		opts+=" --namespace-id= -n"
 			;;
 		"compare")
-		opts+=" --start-block= -s --block-count= -c --data-size= -z \
+		opts+=" --start-block= -s --block-count= -c --block-size= -b --data-size= -z \
 			--metadata-size= -y --ref-tag= -r --data= -d \
 			--metadata= -M --prinfo= -p --app-tag-mask= -m \
 			--app-tag= -a --limited-retry -l \
 			--force-unit-access -f --storage-tag-check -C \
-			--dir-type= -T --dir-spec= -S --dsm= -D --show-command -v \
-			--dry-run -w --latency -t"
+			--dir-type= -T --dir-spec= -S --dsm= -D --show-command -V \
+			--dry-run -w --latency -t --timeout="
 			;;
 		"read")
-		opts+=" --start-block= -s --block-count= -c --data-size= -z \
+		opts+=" --start-block= -s --block-count= -c --block-size= -b --data-size= -z \
 			--metadata-size= -y --ref-tag= -r --data= -d \
 			--metadata= -M --prinfo= -p --app-tag-mask= -m \
 			--app-tag= -a --limited-retry -l \
 			--force-unit-access -f --storage-tag-check -C \
-			--dir-type= -T --dir-spec= -S --dsm= -D --show-command -v \
-			--dry-run -w --latency -t"
+			--dir-type= -T --dir-spec= -S --dsm= -D --show-command -V \
+			--dry-run -w --latency -t --timeout="
 			;;
 		"write")
-		opts+=" --start-block= -s --block-count= -c --data-size= -z \
+		opts+=" --start-block= -s --block-count= -c --block-size= -b --data-size= -z \
 			--metadata-size= -y --ref-tag= -r --data= -d \
 			--metadata= -M --prinfo= -p --app-tag-mask= -m \
 			--app-tag= -a --limited-retry -l \
 			--force-unit-access -f --storage-tag-check -C \
-			--dir-type= -T --dir-spec= -S --dsm= -D --show-command -v \
-			--dry-run -w --latency -t"
+			--dir-type= -T --dir-spec= -S --dsm= -D --show-command -V \
+			--dry-run -w --latency -t --timeout="
 			;;
-		"write-zeros")
+		"write-zeroes")
 		opts+=" --namespace-id= -n --start-block= -s \
 			--block-count= -c --deac -d --limited-retry -l \
 			--force-unit-access -f --prinfo= -p --ref-tag= -r \
 			--app-tag-mask= -m --app-tag= -a \
-			--storage-tag= -S --storage-tag-check -C"
+			--storage-tag= -S --storage-tag-check -C \
+			--dir-type= -T --dir-spec= -S --namespace-zeroes -Z \
+			--timeout= -t"
 			;;
 		"write-uncor")
 		opts+=" --namespace-id= -n --start-block= -s \
-			--block-count= -c"
+			--block-count= -c --dir-type= -T --dir-spec= -S \
+			--timeout= -t"
 			;;
 		"verify")
 		opts+=" --namespace-id= -n --start-block= -s \
 			--block-count= -c --limited-retry -l \
 			--force-unit-access -f --prinfo= -p --ref-tag= -r \
 			--app-tag= -a --app-tag-mask= -m \
-			--storage-tag= -S --storage-tag-check -C"
+			--storage-tag= -S --storage-tag-check -C --timeout= -t"
 			;;
 		"sanitize")
 		opts+=" --no-dealloc -d --oipbp -i --owpass= -n \
-			--ause -u --sanact= -a --ovrpat= -p"
+			--ause -u --sanact= -a --ovrpat= -p --emvs= -e"
+		case $opt in
+			--sanact|-a)
+			vals+=" exit-failure start-block-erase start-overwrite \
+				start-crypto-erase exit-media-verification"
+				;;
+		esac
 			;;
 		"sanitize-log")
 		opts+=" --rae -r --output-format= -o --human-readable -H \
@@ -438,30 +379,30 @@ nvme_list_opts () {
 		opts+=$NO_OPTS
 			;;
 		"show-regs")
-		opts+=" --output-format= -o --human-readable -H"
+		opts+=" --output-format= -o --human-readable -H --timeout= -t"
 			;;
 		"discover")
 		opts+=" --transport= -t -traddr= -a -trsvcid= -s \
-			--host-traddr= -w --hsot-iface= -f \
+			--host-traddr= -w --host-iface= -f \
 			--hostnqn= -q --hostid -I --raw= -r \
 			--raw= -r --device= -d --keep-alive-tmo= -k \
 			--ctrl-loss-tmo= -l --fast-io-fail-tmo= -f \
 			--tos= -T --hdr-digest= -g --data-digest -G \
 			--nr-io-queues= -i --nr-write-queues= -W \
 			--nr-poll-queues= -P --queue-size= -Q \
-			--persistent -p --quiet -S --matching -m \
+			--persistent -p --quiet \
 			--output-format= -o"
 			;;
 		"connect-all")
 		opts+=" --transport= -t -traddr= -a -trsvcid= -s \
-			--host-traddr= -w --hsot-iface= -f \
+			--host-traddr= -w --host-iface= -f \
 			--hostnqn= -q --hostid -I --raw= -r \
 			--raw= -r --device= -d --keep-alive-tmo= -k \
 			--ctrl-loss-tmo= -l --fast-io-fail-tmo= -f \
 			--tos= -T --hdr-digest= -g --data-digest -G \
 			--nr-io-queues= -i --nr-write-queues= -W \
 			--nr-poll-queues= -P --queue-size= -Q \
-			--persistent -p --quiet -S --matching -m \
+			--persistent -p --quiet \
 			--output-format= -o"
 			;;
 		"connect")
@@ -470,9 +411,12 @@ nvme_list_opts () {
 			--nr-poll-queues= -P --queue-size= -Q \
 			--keep-alive-tmo= -k --reconnect-delay= -r \
 			--ctrl-loss-tmo= -l --fast-io-fail-tmo= -f \
-			--tos= -T --duplicate-connect -D --disable-sqflow -d\
+			--tos= -T --duplicate-connect -D --disable-sqflow \
 			--hdr-digest -g --data-digest -G --output-format= -o"
 			;;
+		"dim")
+		opts+=" --task -t --nqn -n --device -d"
+			;;
 		"disconnect")
 		opts+=" --nqn -n --device -d"
 			;;
@@ -485,24 +429,101 @@ nvme_list_opts () {
 		"show-hostnqn")
 		opts+=$NO_OPTS
 			;;
+		"tls-key")
+		opts+=" --output-format= -o --verbose -v --keyring= -k \
+			--keytype= -k --keyfile= -f --import -i \
+			--export -e --revoke= -r"
+			;;
 		"dir-receive")
 		opts+=" --namespace-id= -n --data-len= -l --raw-binary -b \
 			--dir-type= -D --dir-spec= -S --dir-oper= -O \
-			--req-resource= -r --human-readable -H"
+			--req-resource= -r --human-readable -H --timeout= -t"
 			;;
 		"dir-send")
 		opts+=" --namespace-id= -n --data-len= -l --dir-type= -D \
 			--target-dir= -T --dir-spec= -S --dir-oper= -O \
-			--endir= -e --human-readable -H --raw-binary -b"
+			--endir= -e --human-readable -H --raw-binary -b \
+			--timeout= -t"
 			;;
 		"virt-mgmt")
-		opts+=" --cntlid= -c --rt= -r --act= -a --nr= -n"
+		opts+=" --cntlid= -c --rt= -r --act= -a --nr= -n --timeout= -t"
 			;;
 		"rpmb")
 		opts+=" --cmd= -c --msgfile= -f --keyfile= -g \
 			--key= -k --msg= -d --address= -o --blocks= -b \
 			--target= -t"
 			;;
+		"show-topology")
+		opts+=" --output-format= -o --verbose -v --ranking= -r"
+			;;
+		"nvme-mi-recv")
+		opts+=" --opcode= -O --namespace-id= -n --data-len= -l \
+			--nmimt= -m --nmd0= -0 --nmd1= -1 --input-file= -i"
+			;;
+		"nvme-mi-send")
+		opts+=" --opcode= -O --namespace-id= -n --data-len= -l \
+			--nmimt= -m --nmd0= -0 --nmd1= -1 --input-file= -i"
+			;;
+		"get-reg")
+		opts+=" --offset= -O --human-readable -H --cap --vs --cmbloc \
+			--cmbsz --bpinfo --cmbsts --cmbebs --cmbswtp --crto \
+			--pmrcap --pmrsts --pmrebs --pmrswtp --intms --intmc \
+			--cc --csts --nssr --aqa --asq --acq --bprsel --bpmbl \
+			--cmbmsc --nssd --pmrctl --pmrmscl --pmrmscu \
+			--output-format -o --verbose -v --timeout= -t"
+			;;
+		"set-reg")
+		opts+=" --offset= -O --value= -V --mmio32 -m --intms= --intmc= \
+			--cc= --csts= --nssr= --aqa= --asq= --acq= --bprsel= \
+			--bpmbl= --cmbmsc= --nssd= --pmrctl= --pmrmscl= \
+			--pmrmscu= --output-format= -o --verbose -v \
+			--timeout= -t"
+			;;
+		"io-mgmt-recv")
+		opts+=" --namespace-id= -n --mos= -s --mo= -m --data= -d \
+			--data-len= -l --output-format= -o --verbose -v \
+			--timeout= -t"
+			;;
+		"io-mgmt-send")
+		opts+=" --namespace-id= -n --mos= -s --mo= -m --data= -d \
+			--data-len= -l --output-format= -o --verbose -v \
+			--timeout= -t"
+			;;
+		"mgmt-addr-list-log")
+		opts+=" --verbose -v --output-format= -o --timeout= -t"
+			;;
+		"rotational-media-info-log")
+		opts+=" --endg-id= -e --verbose -v --output-format= -o \
+			--timeout= -t"
+			;;
+		"changed-alloc-cns-list-log")
+		opts+=" --output-format= -o --raw-binary -b  --verbose -v \
+			--timeout= -t"
+			;;
+		"dispersed-ns-participating-nss-log")
+		opts+=" --namespace-id= -n --verbose -v --output-format= -o \
+			--timeout= -t"
+			;;
+		"reachability-groups-log")
+		opts+=" --groups-only -g --rae -r --verbose -v \
+			--output-format= -o --timeout= -t"
+			;;
+		"reachability-associations-log")
+		opts+=" --associations-only -a --rae -r --verbose -v \
+			--output-format= -o --timeout= -t"
+			;;
+		"host-discovery-log")
+		opts+=" --all-host-entries -a --rae -r --verbose -v \
+			--output-format= -o --timeout= -t"
+			;;
+		"ave-discovery-log")
+		opts+=" --rae -r --verbose -v --output-format= -o \
+			--timeout= -t"
+			;;
+		"pull--ddc-req-log")
+		opts+=" --rae -r --verbose -v --output-format= -o \
+			--timeout= -t"
+			;;
 		"version")
 		opts+=$NO_OPTS
 			;;
@@ -511,15 +532,19 @@ nvme_list_opts () {
 			;;
 	esac
 
-	opts+=" -h --help"
+	opts+=" -h --help -j --json"
 
-	COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
+	if [[ $vals == " " ]]; then
+		COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
+	else
+		COMPREPLY+=( $( compgen $compargs -W "$vals" -- $val ) )
+	fi
 
 	return 0
 }
 
 plugin_intel_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -575,7 +600,7 @@ plugin_intel_opts () {
 }
 
 plugin_amzn_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -606,70 +631,8 @@ plugin_amzn_opts () {
 	return 0
 }
 
-plugin_lnvm_opts () {
-    local opts=""
-	local compargs=""
-
-	local nonopt_args=0
-	for (( i=0; i < ${#words[@]}-1; i++ )); do
-		if [[ ${words[i]} != -* ]]; then
-			let nonopt_args+=1
-		fi
-	done
-
-	if [ $nonopt_args -eq 3 ]; then
-		opts="/dev/nvme* "
-	fi
-
-	opts+=" "
-
-	case "$1" in
-		"list")
-		opts+=$NO_OPTS
-			;;
-		"info")
-		opts+=$NO_OPTS
-			;;
-		"id-ns")
-		opts+=" --namespace-id= -n --raw-binary -b --human-readable -H"
-			;;
-		"chunk-log")
-		opts+=" --output-format= -o --human-readable -H"
-			;;
-		"init")
-		opts+=" --device-name= -d --mediamgr-name -m"
-			;;
-		"create")
-		opts+=" --device-name= -d --target-name= -n --target-type= -t \
-			--lun-begin= -b --lun-end= -e --over-prov= -o --factory -f"
-			;;
-		"remove")
-		opts+=" --target-name= -n"
-			;;
-		"factory")
-		opts+=" --device-name= -d --erase-only-marked -e
-			--clear-host-side-blks -s --clear-bb-blks -b"
-			;;
-		"diag-bbtbl")
-		opts+=" --namespace-id= -n --channel-id= -c --lun-id= -l \
-			--raw-binary -b"
-			;;
-		"diag-set-bbtbl")
-		opts+=" --namespace-id= -n --channel-id= -c --lun-id= -l \
-			--plane-id= -p --block-id= -b --value= -v"
-			;;
-		"help")
-		opts+=$NO_OPTS
-			;;
-	esac
-
-	COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
-
-	return 0
-}
-
 plugin_memblaze_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -724,7 +687,7 @@ plugin_memblaze_opts () {
 }
 
 plugin_wdc_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -836,7 +799,7 @@ plugin_wdc_opts () {
 }
 
 plugin_huawei_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -871,7 +834,7 @@ plugin_huawei_opts () {
 }
 
 plugin_toshiba_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -908,7 +871,7 @@ plugin_toshiba_opts () {
 }
 
 plugin_micron_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -983,8 +946,8 @@ plugin_micron_opts () {
 	return 0
 }
 
-lugin_seagate_opts () {
-    local opts=""
+plugin_seagate_opts () {
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1039,7 +1002,7 @@ lugin_seagate_opts () {
 }
 
 plugin_virtium_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1073,7 +1036,7 @@ plugin_virtium_opts () {
 }
 
 plugin_shannon_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1098,7 +1061,7 @@ plugin_shannon_opts () {
 			--data-len= -l --raw-binary -b --cdw11= -c --human-readable -H"
 			;;
 		"set-feature-add")
-		opts+=" --namespace-id= -n --feature-id= -f --value= -v \
+		opts+=" --namespace-id= -n --feature-id= -f --value= -V \
 			--data-len= -l --data= -d --save -s"
 			;;
 		"id-ctrl")
@@ -1116,7 +1079,7 @@ plugin_shannon_opts () {
 }
 
 plugin_dera_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1147,7 +1110,7 @@ plugin_dera_opts () {
 }
 
 plugin_sfx_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1195,8 +1158,133 @@ plugin_sfx_opts () {
 	return 0
 }
 
+plugin_solidigm_opts () {
+	local opts=""
+	local compargs=""
+
+	local nonopt_args=0
+	for (( i=0; i < ${#words[@]}-1; i++ )); do
+		if [[ ${words[i]} != -* ]]; then
+			let nonopt_args+=1
+		fi
+	done
+
+	if [ $nonopt_args -eq 3 ]; then
+		opts="/dev/nvme* "
+	fi
+
+	opts+=" "
+
+	case "$1" in
+		"id-ctrl")
+		opts+=" --raw-binary -b --human-readable -H \
+			--vendor-specific -v --output-format= -o \
+			--verbose -v "
+			;;
+		"vs-smart-add-log")
+		opts+="--output-format= -o"
+			;;
+		"garbage-collect-log")
+		opts+="--output-format= -o"
+			;;
+		"vs-internal-log")
+		opts+=" --type= -t --namespace-id= -n \
+		--file-prefix= -p --verbose -v"
+			;;
+		"latency-tracking-log")
+		opts+=" --enable -e --disable -d \
+		--read -r --write -w \
+		--type -t --output-format -o"
+			;;
+		"clear-pcie-correctable-errors")
+		opts+=" --no-uuid -n"
+			;;
+		"parse-telemetry-log")
+		opts+=" --host-generate -g --controller-init -c \
+		--data-area -d --config-file -j \
+		--source-file -s"
+			;;
+		"clear-fw-activate-history")
+		opts+=" --no-uuid -n"
+			;;
+		"vs-fw-activate-history")
+		opts+=" --output-format -o"
+			;;
+		"log-page-directory")
+		opts+=" --output-format -o"
+			;;
+		"vs-drive-info")
+		opts+=" "
+			;;
+		"cloud-SSDplugin-version")
+		opts+=$NO_OPTS
+			;;
+		"market-log")
+		opts+=" --raw-binary -b"
+			;;
+		"smart-log-add")
+		opts+=" --namespace-id= -n --output-format -o"
+			;;
+		"temp-stats")
+		opts+=" --raw-binary -b"
+			;;
+		"workload-tracker")
+		opts+=" --uuid-index= -U --enable -e --disable -d --sample-time= -s \
+		--type= -t --run-time= -r --flush-freq= -f \
+		--wall-clock -w --trigger-field= -T \
+		--trigger-threshold= -V --trigger-on-delta -D \
+		--trigger-on-latency -L --verbose -v"
+			;;
+		"version")
+		opts+=$NO_OPTS
+			;;
+		"help")
+		opts+=$NO_OPTS
+			;;
+	esac
+
+	COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
+
+	return 0
+}
+
+plugin_fdp_opts () {
+	local opts=""
+	local compargs=""
+
+	local nonopt_args=0
+	for (( i=0; i < ${#words[@]}-1; i++ )); do
+		if [[ ${words[i]} != -* ]]; then
+			let nonopt_args+=1
+		fi
+	done
+
+	if [ $nonopt_args -eq 3 ]; then
+		opts="/dev/nvme* "
+	fi
+
+	opts+=" "
+
+	case "$1" in
+		"feature")
+		opts+=" --endgrp-id= -e --enable-conf-idx= -c \
+		--disable -d --verbose -v"
+			;;
+		"version")
+		opts+=$NO_OPTS
+			;;
+		"help")
+		opts+=$NO_OPTS
+			;;
+	esac
+
+	COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
+
+	return 0
+}
+
 plugin_transcend_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1229,8 +1317,40 @@ plugin_transcend_opts () {
 	return 0
 }
 
+plugin_dapustor_opts () {
+	local opts=""
+	local compargs=""
+
+	local nonopt_args=0
+	for (( i=0; i < ${#words[@]}-1; i++ )); do
+		if [[ ${words[i]} != -* ]]; then
+			let nonopt_args+=1
+		fi
+	done
+
+	if [ $nonopt_args -eq 3 ]; then
+		opts="/dev/nvme* "
+	fi
+
+	opts+=" "
+
+	case "$1" in
+		"smart-log-add")
+		opts+=" --namespace-id= -n --raw-binary -b \
+			--json -j"
+			;;
+		"help")
+		opts+=$NO_OPTS
+			;;
+	esac
+
+	COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
+
+	return 0
+}
+
 plugin_zns_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1261,7 +1381,7 @@ plugin_zns_opts () {
 			--data-len= -l"
 			;;
 		"zone-mgmt-send")
-		opts+=" --namespace-id= -n --start-lba= -s \
+		opts+=" --namespace-id= -n --start-lba= -s --zsaso -o \
 			--select-all -a --zsa= -z --data-len= -l \
 			--data= -d --timeout= -t"
 			;;
@@ -1280,7 +1400,7 @@ plugin_zns_opts () {
 			;;
 		"open-zone")
 		opts+=" --namespace-id= -n --start-lba= -s \
-			--select-all -a --timeout= -t"
+			--select-all -a --timeout= -t --zrwa -r"
 			;;
 		"reset-zone")
 		opts+=" --namespace-id= -n --start-lba= -s \
@@ -1292,7 +1412,10 @@ plugin_zns_opts () {
 			;;
 		"set-zone-desc")
 		opts+=" --namespace-id= -n --start-lba= -s \
-			--data= -d --timeout= -t"
+			--data= -d --timeout= -t --zrwa -r"
+			;;
+		"flush-zone")
+		opts+=" --namespace-id= -n --last-lba= -l --timeout= -t"
 			;;
 		"zone-append")
 		opts+=" --namespace-id= -n --zslba= -s --data-size= -z \
@@ -1315,7 +1438,7 @@ plugin_zns_opts () {
 }
 
 plugin_nvidia_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1347,7 +1470,7 @@ plugin_nvidia_opts () {
 }
 
 plugin_ymtc_opts () {
-    local opts=""
+	local opts=""
 	local compargs=""
 
 	local nonopt_args=0
@@ -1368,7 +1491,177 @@ plugin_ymtc_opts () {
 		opts+=" --namespace-id= -n --raw-binary -b"
 			;;
 		"help")
-		opts+=NO_OPTS
+		opts+=$NO_OPTS
+			;;
+	esac
+
+	COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
+
+	return 0
+}
+
+plugin_inspur_opts () {
+	local opts=""
+	local compargs=""
+
+	local nonopt_args=0
+	for (( i=0; i < ${#words[@]}-1; i++ )); do
+		if [[ ${words[i]} != -* ]]; then
+			let nonopt_args+=1
+		fi
+	done
+
+	if [ $nonopt_args -eq 3 ]; then
+		opts="/dev/nvme* "
+	fi
+
+	opts+=" "
+
+	case "$1" in
+		"nvme-vendor-log")
+		opts+=$NO_OPTS
+			;;
+		"help")
+		opts+=$NO_OPTS
+			;;
+	esac
+
+	COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
+
+	return 0
+}
+
+plugin_mangoboost_opts () {
+	local opts=""
+	local compargs=""
+
+	local nonopt_args=0
+	for (( i=0; i < ${#words[@]}-1; i++ )); do
+		if [[ ${words[i]} != -* ]]; then
+			let nonopt_args+=1
+		fi
+	done
+
+	if [ $nonopt_args -eq 3 ]; then
+		opts="/dev/nvme* "
+	fi
+
+	opts+=" "
+
+	case "$1" in
+		"id-ctrl")
+		opts+=" --raw-binary -b --human-readable -H \
+			--vendor-specific -v --output-format= -o"
+			;;
+		"help")
+		opts+=$NO_OPTS
+			;;
+	esac
+
+	COMPREPLY+=( $( compgen $compargs -W "$opts" -- $cur ) )
+
+	return 0
+}
+
+plugin_ocp_opts () {
+	local opts=""
+	local compargs=""
+
+	local nonopt_args=0
+	for (( i=0; i < ${#words[@]}-1; i++ )); do
+		if [[ ${words[i]} != -* ]]; then
+			let nonopt_args+=1
+		fi
+	done
+
+	if [ $nonopt_args -eq 3 ]; then
+		opts="/dev/nvme* "
+	fi
+
+	opts+=" "
+
+	case "$1" in
+		"smart-add-log")
+		opts+=" --output-format= -o"
+			;;
+		"latency-monitor-log")
+		opts+=" --output-format= -o"
+			;;
+		"set-latency-monitor-feature")
+		opts+=" --active_bucket_timer_threshold= -t \
+			--active_threshold_a= -a --active_threshold_b= -b \
+			--active_threshold_c= -c --active_threshold_d= -d \
+			--active_latency_config= -f \
+			--active_latency_minimum_window= -w \
+			--debug_log_trigger_enable -r --discard_debug_log= -l \
+			--latency_monitor_feature_enable= -e"
+			;;
+		"internal-log")
+		opts+=" --telemetry-log= -l --string-log= -s \
+			--output-file= -f --output-format= -o \
+			--data-area= -a --telemetry-type= -t"
+			;;
+		"clear-fw-activate-history")
+		opts+=" --no-uuid -n"
+			;;
+		"eol-plp-failure-mode")
+		opts+=" --mode= -m --save -s --sel= -S --no-uuid -n"
+			;;
+		"clear-pcie-correctable-error-counters")
+		opts+=" --no-uuid -n"
+			;;
+		"fw-activate-history")
+		opts+=" --output-format= -o"
+			;;
+		"device-capability-log")
+		opts+=" --output-format= -o"
+			;;
+		"set-dssd-power-state-feature")
+		opts+=" --power-state= -p --no-uuid -n --save -s"
+			;;
+		"get-dssd-power-state-feature")
+		opts+=" --sel= -S --all -a --no-uuid -n"
+			;;
+		"telemetry-string-log")
+		opts+=" --output-file= -f --output-format= -o"
+			;;
+		"set-telemetry-profile")
+		opts+=" --telemetry-profile-select= -t"
+			;;
+		"set-dssd-async-event-config")
+		opts+=" --enable-panic-notices -e --save -s"
+			;;
+		"get-dssd-async-event-config")
+		opts+=" --sel= -S"
+			;;
+		"tcg-configuration-log")
+		opts+=" --output-file= -o"
+			;;
+		"get-error-injection")
+		opts+=" --sel= -s --no-uuid -n --all-ns -a"
+			;;
+		"set-error-injection")
+		opts+=" --data= -d --number= -n --no-uuid -N --type= -t \
+			--nrtdp= -r --verbose -v --output-format -o --timeout= --all-ns -a"
+			;;
+		"hardware-component-log")
+		opts+=" --comp-id= -i --list -l --verbose -v \
+			--output-format -o --timeout= -t"
+			;;
+		"get-latency-monitor")
+		opts+=" --sel= -s \
+			--namespace-id= -n --no-uuid -u"
+			;;
+		"get-clear-pcie-correctable-errors")
+		opts+=" --sel= -s \
+			--namespace-id= -n --no-uuid -u"
+			;;
+		"get-telemetry-profile")
+		opts+=" --sel= -s \
+			--namespace-id= -n --no-uuid -u"
+			;;
+		"help")
+		opts+=$NO_OPTS
 			;;
 	esac
 
@@ -1381,6 +1674,143 @@ _nvme_subcmds () {
 	local cur prev words cword
 	_init_completion || return
 
+	# Constant to indicate command has no options
+	NO_OPTS=""
+
+	# Associative array of plugins and associated subcommands
+	# Order here is same as PLUGIN_OBJS in Makefile
+	typeset -Ar _plugin_subcmds=(
+		[intel]="id-ctrl internal-log lat-stats \
+			set-bucket-thresholds lat-stats-tracking \
+			market-name smart-log-add temp-stats"
+		[amzn]="id-ctrl"
+		[memblaze]="smart-log-add get-pm-status set-pm-status \
+			select-download lat-stats lat-stats-print lat-log \
+			lat-log-print clear-error-log"
+		[wdc]="cap-diag drive-log get-crash-dump get-pfail-dump \
+			id-ctrl purge purge-monitor vs-internal-log \
+			vs-nand-stats vs-smart-add-log clear-pcie-correctable-errors \
+			drive-essentials get-drive-status clear-assert-dump \
+			drive-resize vs-fw-activate-history clear-fw-activate-history \
+			enc-get-log vs-telemetry-controller-option \
+			vs-error-reason-identifier log-page-directory \
+			namespace-resize vs-drive-info vs-temperature-stats \
+			capabilities cloud-SSD-plugin-version vs-pcie-stats"
+		[huawei]="list id-ctrl"
+		[netapp]="smdevices ontapdevices"
+		[toshiba]="vs-smart-add-log vs-internal-log \
+			clear-pcie-correctable-errors"
+		[micron]="select-download vs-temperature-stats vs-pcie-stats \
+			clear-pcie-correctable-errors vs-internal-log \
+			vs-telemetry-controller-option vs-nand-stats \
+			vs-drive-info plugin-version cloud-SSD-plugin-version \
+			log-page-directory vs-fw-activate-history \
+			vs-error-reason-identifier vs-smart-add-log \
+			clear-fw-activate-history vs-smbus-option ocp-telemetry-log-parse"
+		[seagate]="vs-temperature-stats vs-log-page-sup \
+			vs-smart-add-log vs-pcie-stats clear-pcie-correctable-errors \
+			get-host-tele get-ctrl-tele vs-internal-log \
+			plugin-version"
+		[virtium]="save-smart-to-vtview-log show-identify"
+		[shannon]="smart-log-add get-feature-add set-feature-add id-ctrl"
+		[dera]="smart-log-add"
+		[sfx]="smart-log-add lat-stats get-bad-block query-cap \
+			change-cap set-feature get-feature"
+		[solidigm]="id-ctrl vs-smart-add-log garbage-collect-log \
+			vs-internal-log latency-tracking-log \
+			clear-pcie-correctable-errors parse-telemetry-log \
+			clear-fw-activate-history vs-fw-activate-history log-page-directory \
+			vs-drive-info cloud-SSDplugin-version market-log \
+			smart-log-add temp-stats workload-tracker version help"
+		[fdp]="feature version help"
+		[transcend]="healthvalue badblock"
+		[dapustor]="smart-log-add"
+		[zns]="id-ctrl id-ns zone-mgmt-recv \
+			zone-mgmt-send report-zones close-zone \
+			finish-zone open-zone reset-zone offline-zone \
+			set-zone-desc zone-append changed-zone-list"
+		[nvidia]="id-ctrl"
+		[ymtc]="smart-log-add"
+		[inspur]="nvme-vendor-log"
+		[ocp]="smart-add-log latency-monitor-log \
+			set-latency-monitor-feature internal-log \
+			clear-fw-activate-history eol-plp-failure-mode \
+			clear-pcie-correctable-error-counters \
+			vs-fw-activate-history device-capability-log \
+			set-dssd-power-state-feature get-dssd-power-state-feature \
+			telemetry-string-log set-telemetry-profile \
+			set-dssd-async-event-config get-dssd-async-event-config \
+			get-error-injection set-error-injection \
+			hardware-component-log get-latency-monitor \
+			get-clear-pcie-correctable-errors get-telemetry-profile"
+		[mangoboost]="id-ctrl"
+	)
+
+	# Associative array mapping plugins to corresponding option completions
+	typeset -Ar _plugin_funcs=(
+		[intel]="plugin_intel_opts"
+		[amzn]="plugin_amzn_opts"
+		[memblaze]="plugin_memblaze_opts"
+		[wdc]="plugin_wdc_opts"
+		[huawei]="plugin_huawei_opts"
+		[toshiba]="plugin_toshiba_opts"
+		[micron]="plugin_micron_opts"
+		[seagate]="plugin_seagate_opts"
+		[virtium]="plugin_virtium_opts"
+		[shannon]="plugin_shannon_opts"
+		[dera]="plugin_dera_opts"
+		[sfx]="plugin_sfx_opts"
+		[solidigm]="plugin_solidigm_opts"
+		[fdp]="plugin_fdp_opts"
+		[transcend]="plugin_transcend_opts"
+		[dapustor]="plugin_dapustor_opts"
+		[zns]="plugin_zns_opts"
+		[nvidia]="plugin_nvidia_opts"
+		[ymtc]="plugin_ymtc_opts"
+		[inspur]="plugin_inspur_opts"
+		[ocp]="plugin_ocp_opts"
+		[mangoboost]="plugin_mangoboost_opts"
+	)
+
+	# Top level commands
+	_cmds="list list-subsys id-ctrl id-ns \
+		id-ns-granularity list-ns list-ctrl \
+		id-ns-lba-format nvm-id-ns nvm-id-ns-lba-format \
+		nvm-id-ctrl primary-ctrl-caps list-secondary \
+		ns-descs id-nvmset id-uuid id-iocs id-domain create-ns \
+		delete-ns get-ns-id get-log telemetry-log \
+		fw-log changed-ns-list-log smart-log ana-log \
+		error-log effects-log endurance-log \
+		predictable-lat-log pred-lat-event-agg-log \
+		persistent-event-log endurance-agg-log \
+		lba-status-log resv-notif-log get-feature \
+		device-self-test self-test-log set-feature \
+		set-property get-property format fw-commit \
+		fw-download admin-passthru io-passthru \
+		security-send security-recv get-lba-status \
+		resv-acquire resv-register resv-release \
+		resv-report dsm copy flush compare read \
+		write write-zeroes write-uncor verify \
+		sanitize sanitize-log reset subsystem-reset \
+		ns-rescan show-regs discover connect-all \
+		connect disconnect disconnect-all gen-hostnqn \
+		show-hostnqn tls-key dir-receive dir-send virt-mgmt \
+		rpmb boot-part-log fid-support-effects-log \
+		supported-log-pages lockdown media-unit-stat-log \
+		supported-cap-config-log dim show-topology list-endgrp \
+		nvme-mi-recv nvme-mi-send get-reg set-reg mgmt-addr-list-log \
+		rotational-media-info-log changed-alloc-ns-list-log \
+		io-mgmt-recv io-mgmt-send dispersed-ns-participating-nss-log \
+		reachability-groups-log reachability-associations-log \
+		host-discovery-log ave-discovery-log pull--ddc-req-log"
+
+	# Add plugins:
+	for plugin in "${!_plugin_subcmds[@]}"; do
+		_cmds+=" $plugin"
+	done
+
+	_cmds+=" version help"
+
 	if [[ ${#words[*]} -lt 3 ]]; then
 		COMPREPLY+=( $(compgen -W "$_cmds" -- $cur ) )
 	else
